from __future__ import annotations

import sys
from pathlib import Path
from typing import TYPE_CHECKING

from PyQt5 import QtCore, QtGui, QtWidgets

import ztrack.gui

if TYPE_CHECKING:
    from typing import Type


def launch(
    Widget: Type[QtWidgets.QWidget],
    style: str = "dark",
    modern_window=False,
    windowState=QtCore.Qt.WindowMaximized,
    **kwargs,
) -> int:
    try:
        import ctypes

        ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID("ztrack")
    except ModuleNotFoundError:
        pass
    except AttributeError:
        pass

    app = QtWidgets.QApplication(sys.argv)
    widget = Widget(**kwargs)
    icon_path = str(Path(ztrack.gui.__file__).parent / "img" / "logo.svg")
    app.setWindowIcon(QtGui.QIcon(icon_path))

    try:
        import qtmodern.styles
        import qtmodern.windows

        if modern_window:
            widget = qtmodern.windows.ModernWindow(widget)
        getattr(qtmodern.styles, style)(app)
    except TypeError:
        pass
    except ModuleNotFoundError:
        app.setStyle(style)  # type: ignore
    except AttributeError:
        app.setStyle(style)  # type: ignore
    finally:
        widget.setWindowState(windowState)
        widget.setVisible(True)
        return app.exec()



from abc import ABC, abstractmethod
from typing import Type

import numpy as np
import pandas as pd
from decord import VideoReader
from tqdm import tqdm

from ztrack.utils.exception import TrackingError
from ztrack.utils.variable import Rect

from .params import Params


class Tracker(ABC):
    def __init__(self, roi=None, params: dict = None, *, verbose=0):
        self._roi = Rect("", roi)
        self._params = self._Params(params)
        self._verbose = verbose

    def __repr__(self):
        return f"{self.__class__.__name__}(roi={self._roi.value}, params={self.params.to_dict()})"

    @property
    @abstractmethod
    def _Params(self) -> Type[Params]:
        pass

    def _get_bbox_img(self, frame: np.ndarray):
        return frame[self._roi.to_slice()]

    @property
    def roi(self):
        return self._roi

    @roi.setter
    def roi(self, bbox):
        self._roi = bbox

    @property
    @abstractmethod
    def shapes(self):
        pass

    def annotate(self, frame: np.ndarray) -> None:
        try:
            img = self._get_bbox_img(frame)
            results = self._results_to_series(self._track_img(img))
            self.annotate_from_series(results)
        except TrackingError:
            for shape in self.shapes:
                shape.visible = False

    @abstractmethod
    def annotate_from_series(self, s: pd.Series) -> None:
        pass

    @property
    def params(self) -> Params:
        return self._params

    @staticmethod
    @abstractmethod
    def name():
        pass

    @staticmethod
    @abstractmethod
    def display_name():
        pass

    def _track_frame(self, frame: np.ndarray) -> pd.Series:
        results = self._track_img(frame[self.roi.to_slice()])
        return self._results_to_series(
            self._transform_from_roi_to_frame(results)
        )

    @classmethod
    @abstractmethod
    def _results_to_series(cls, results):
        pass

    @abstractmethod
    def _transform_from_roi_to_frame(self, results):
        pass

    @abstractmethod
    def _track_img(self, img: np.ndarray):
        pass

    def track_video(self, video_path):
        video_reader = VideoReader(str(video_path))
        it = (
            tqdm(range(len(video_reader)))
            if self._verbose
            else range(len(video_reader))
        )
        return pd.DataFrame(
            [self._track_frame(video_reader[i].asnumpy()) for i in it]
        )



from __future__ import annotations

import json
from typing import TYPE_CHECKING

from PyQt5 import QtWidgets

from ztrack._settings import config_extension
from ztrack.gui.utils.file import selectVideoDirectories, selectVideoPaths
from ztrack.tracking import get_trackers
from ztrack.utils.file import get_config_dict, get_paths_for_config_creation

from ._control_widget import ControlWidget
from ._main_window import MainWindow

if TYPE_CHECKING:
    from typing import List, Optional

    from PyQt5 import QtGui

    from ztrack.tracking.tracker import Tracker


class CreateConfigWindow(MainWindow):
    def __init__(
        self,
        parent: QtWidgets.QWidget = None,
        videoPaths: List[str] = None,
        savePaths: List[List[str]] = None,
        verbose=False,
    ):
        super().__init__(parent, videoPaths=videoPaths, verbose=verbose)

        if savePaths is None:
            savePaths = []

        self._savePaths: List[List[str]] = savePaths

        self._trackerGroups = get_trackers()
        self._controlWidget = ControlWidget(self)

        self._buttonBox = QtWidgets.QDialogButtonBox(self)
        self._buttonBox.setStandardButtons(
            QtWidgets.QDialogButtonBox.Ok  # type: ignore
            | QtWidgets.QDialogButtonBox.Cancel
        )

        self._hBoxLayout.addWidget(self._controlWidget)
        self._hBoxLayout.setStretch(0, 50)
        self._hBoxLayout.setStretch(1, 50)
        self._layout.addWidget(self._buttonBox)

        for k, v in self._trackerGroups.items():
            self._addTrackerGroup(k, v)

        self._trackingPlotWidget.setTrackerGroup(list(self._trackerGroups)[0])

        self._controlWidget.currentChanged.connect(self._onTabChanged)
        self._controlWidget.trackerChanged.connect(self._onTrackerChanged)
        self._controlWidget.paramsChanged.connect(self._onParamsChanged)
        self._trackingPlotWidget.roiChanged.connect(self._onRoiChanged)

        self._buttonBox.button(
            QtWidgets.QDialogButtonBox.Cancel
        ).clicked.connect(self._onCancelButtonClicked)

        self._buttonBox.button(QtWidgets.QDialogButtonBox.Ok).clicked.connect(
            self._onOkButtonClicked
        )
        self.updateVideo()

    @property
    def trackingPlotWidget(self):
        return self._trackingPlotWidget

    @property
    def _currentSavePaths(self) -> Optional[List[str]]:
        if len(self._savePaths) > 0:
            return self._savePaths[0]

        return None

    def _saveTrackingConfig(self):
        trackingConfig = {}

        for group_name, trackers in self._trackerGroups.items():
            tracker = trackers[
                self._controlWidget.getCurrentTrackerIndex(group_name)
            ]
            trackingConfig[group_name] = dict(
                method=tracker.name(),
                roi=tracker.roi.value,
                params=tracker.params.to_dict(),
            )

        for savePath in self._currentSavePaths:
            with open(savePath + config_extension, "w") as fp:
                json.dump(trackingConfig, fp)

    def _onOkButtonClicked(self):
        self._saveTrackingConfig()
        self.dequeue()
        self.updateVideo()

    def _onCancelButtonClicked(self):
        self.dequeue()
        self.updateVideo()

    def dropEvent(self, event: QtGui.QDropEvent) -> None:
        paths = [u.toLocalFile() for u in event.mimeData().urls()]

        for path in paths:
            self.enqueue(path, [path], first=True)

        self.updateVideo()

    def _onFrameChanged(self):
        img = self._currentFrame

        if img is not None:
            self._trackingPlotWidget.setImage(img)

            for name, tracker in self._trackerGroups.items():
                index = self._controlWidget.getCurrentTrackerIndex(name)
                tracker[index].annotate(img)
                self._trackingPlotWidget.updateRoiGroups()

    def _onTrackerChanged(self, name: str, index: int):
        self._trackingPlotWidget.setTracker(name, index)
        img = self._currentFrame

        if img is not None:
            self._trackerGroups[name][index].annotate(self._currentFrame)
            self._trackingPlotWidget.updateRoiGroups()

    def _onRoiChanged(self, name: str):
        img = self._currentFrame

        if img is not None:
            index = self._controlWidget.getCurrentTrackerIndex(name)
            self._trackerGroups[name][index].annotate(img)
            self._trackingPlotWidget.updateRoiGroups()

    def _onTabChanged(self, index: int):
        name = list(self._trackerGroups)[index]
        self._trackingPlotWidget.setTrackerGroup(name)

    def _onParamsChanged(self, name: str, index: int):
        img = self._currentFrame

        if img is not None:
            self._trackerGroups[name][index].annotate(img)
            self._trackingPlotWidget.updateRoiGroups()

    def _addTrackerGroup(self, name: str, trackers: List[Tracker]):
        self._controlWidget.addTrackerGroup(name, trackers)
        self._trackingPlotWidget.addTrackerGroup(name, trackers)

    def _setStateFromTrackingConfig(self, trackingConfig: dict):
        self._controlWidget.setStateFromTrackingConfig(trackingConfig)
        self._trackingPlotWidget.setStateFromTrackingConfig(trackingConfig)

    def updateVideo(self):
        if self._currentVideoPath is not None:
            trackingConfig = get_config_dict(self._currentVideoPath)

            if trackingConfig is not None:
                self._setStateFromTrackingConfig(trackingConfig)

        super().updateVideo()

    def enqueue(self, videoPath: str, savePaths: List[str], first=False):
        if first:
            self._videoPaths.insert(0, videoPath)
            self._savePaths.insert(0, savePaths)
        else:
            self._videoPaths.append(videoPath)
            self._savePaths.append(savePaths)

    def dequeue(self):
        if len(self._videoPaths) > 0:
            self._videoPaths.pop(0)
            self._savePaths.pop(0)

    def _openFiles(self):
        videoPaths = selectVideoPaths(native=True)

        for videoPath in reversed(videoPaths):
            self.enqueue(videoPath, [videoPath], first=True)

        self.updateVideo()

    def _openFolders(self):
        directories, (
            recursive,
            sameConfig,
            overwrite,
        ) = selectVideoDirectories(
            (
                ("Include subdirectories", True),
                ("Use one configuration file per directory", True),
                ("Overwrite existing configuration files", True),
            )
        )
        videoPaths, savePaths = get_paths_for_config_creation(
            directories, recursive, sameConfig, overwrite
        )

        for videoPath, savePath in zip(videoPaths, savePaths):
            self.enqueue(videoPath, savePath)

        self.updateVideo()



import cv2
import numpy as np
from skimage.draw import circle_perimeter

from ztrack.utils.exception import TrackingError
from ztrack.utils.geometry import angle_diff
from ztrack.utils.math import split_int
from ztrack.utils.variable import Angle, Float, Int, Point

from .tail_tracker import TailParams, TailTracker


class EmbeddedTailTracker(TailTracker):
    class __Params(TailParams):
        def __init__(self, params: dict = None):
            super().__init__(params)
            self.sigma = Float("Sigma (px)", 0, 0, 100, 0.1)
            self.n_steps = Int("Number of steps", 10, 3, 20)
            self.n_points = Int(
                "Number of points", 7, 2, EmbeddedTailTracker.max_n_points
            )
            self.length = Int("Tail length (px)", 200, 0, 1000)
            self.tail_base = Point("Tail base (x, y)", (250, 120))
            self.angle = Angle("Initial angle (°)", 90)
            self.theta = Angle("Search angle (°)", 60)

    def __init__(self, roi=None, params: dict = None, *, verbose=0):
        super().__init__(roi, params, verbose=verbose)

    @property
    def _Params(self):
        return self.__Params

    @staticmethod
    def _preprocess(img, sigma=0):
        img = cv2.cvtColor(img, cv2.COLOR_RGB2GRAY)

        if cv2.mean(img)[0] > 127:
            img = cv2.bitwise_not(img)

        if sigma > 0:
            img = cv2.GaussianBlur(img, (0, 0), sigma)

        return img

    def _track_tail(self, img):
        x, y = self.params.tail_base
        if self.roi.value is not None:
            x0, y0 = self.roi.value[:2]
            point = (x - x0, y - y0)
        else:
            point = (x, y)
        angle = np.deg2rad(self.params.angle)
        theta = np.deg2rad(self.params.theta / 2)
        img = self._preprocess(img, self.params.sigma)
        h, w = img.shape
        tail = np.zeros((self.params.n_steps + 1, 2), dtype=int)
        tail[0] = point
        step_lengths = split_int(
            round(self.params.length), self.params.n_steps
        )
        for i in range(self.params.n_steps):
            points = np.column_stack(
                circle_perimeter(*point, step_lengths[i], shape=(w, h))
            )
            angles = np.arctan2(*reversed((points - point).T))
            idx = angle_diff(angles, angle) < theta
            points, angles = points[idx], angles[idx]
            x, y = points.T

            try:
                argmax = img[y, x].argmax()
            except ValueError:
                raise TrackingError

            angle = angles[argmax]
            tail[i + 1] = point = points[argmax]
        return tail

    @staticmethod
    def name():
        return "embedded"

    @staticmethod
    def display_name():
        return "Embedded"



from typing import Dict, Iterable, List, Type

from .eye import trackers as eye_trackers
from .tail import trackers as tail_trackers
from .tracker import Tracker

_trackers: Dict[str, Iterable[Type[Tracker]]] = {
    "eye": eye_trackers,
    "tail": tail_trackers,
}

_trackers_dict: Dict[str, Dict[str, Type[Tracker]]] = {
    group_name: {tracker.name(): tracker for tracker in trackers}
    for group_name, trackers in _trackers.items()
}


def get_trackers() -> Dict[str, List[Tracker]]:
    return {key: [i() for i in value] for key, value in _trackers.items()}


def get_trackers_from_config(config_dict, **kwargs) -> Dict[str, Tracker]:
    return {
        group_name: _trackers_dict[group_name][group_dict["method"]](
            group_dict["roi"], group_dict["params"], **kwargs
        )
        for group_name, group_dict in config_dict.items()
    }



from abc import ABC, abstractmethod

import cv2
import numpy as np
import pandas as pd

from ztrack.tracking.params import Params
from ztrack.tracking.tracker import Tracker
from ztrack.utils.cv import binary_threshold, find_contours
from ztrack.utils.geometry import wrap_degrees
from ztrack.utils.shape import Ellipse


class EyeParams(Params, ABC):
    pass


class EyeTracker(Tracker, ABC):
    _index = pd.MultiIndex.from_product(
        (
            ("left_eye", "right_eye", "swim_bladder"),
            ("cx", "cy", "a", "b", "theta"),
        )
    )

    def __init__(self, roi=None, params: dict = None, *, verbose=0):
        super().__init__(roi, params, verbose=verbose)
        self._left_eye = Ellipse(0, 0, 1, 1, 0, 4, "b")
        self._right_eye = Ellipse(0, 0, 1, 1, 0, 4, "r")
        self._swim_bladder = Ellipse(0, 0, 1, 1, 0, 4, "g")

    @property
    def shapes(self):
        return [self._left_eye, self._right_eye, self._swim_bladder]

    @abstractmethod
    def _track_ellipses(self, src: np.ndarray) -> np.ndarray:
        pass

    @staticmethod
    def _preprocess(img, sigma=0):
        img = cv2.cvtColor(img, cv2.COLOR_RGB2GRAY)

        if cv2.mean(img)[0] > 127:
            img = cv2.bitwise_not(img)

        if sigma > 0:
            img = cv2.GaussianBlur(img, (0, 0), sigma)

        return img

    @staticmethod
    def _binary_segmentation(img, threshold):
        return find_contours(binary_threshold(img, threshold))

    @staticmethod
    def _correct_orientation(ellipses):
        centers = ellipses[:, :2]
        midpoint = centers[:2].mean(0)
        midline = midpoint - centers[2]
        heading = np.rad2deg(np.arctan2(*midline[::-1]))
        is_opposite = abs(wrap_degrees(heading - ellipses[:, -1])) > 90
        ellipses[is_opposite, -1] = wrap_degrees(
            ellipses[is_opposite, -1] - 180
        )
        return ellipses

    @staticmethod
    def _fit_ellipses(contours):
        ellipses = np.array(
            [
                (x, y, b / 2, a / 2, theta - 90)
                for (x, y), (a, b), theta in map(
                    cv2.fitEllipse, map(cv2.convexHull, contours)
                )
            ]
        )

        return ellipses

    @staticmethod
    def _sort_centers(centers):
        idx = np.arange(3)
        swim_bladder = np.argmin(
            [np.linalg.norm(np.subtract(*centers[idx != i])) for i in idx]
        )
        eyes = idx[idx != swim_bladder]
        left_eye, right_eye = (
            eyes
            if np.cross(*centers[eyes] - centers[swim_bladder]) > 0
            else eyes[::-1]
        )
        return left_eye, right_eye, swim_bladder

    def annotate_from_series(self, series: pd.Series) -> None:
        ellipse_shapes = [self._left_eye, self._right_eye, self._swim_bladder]
        body_parts = ["left_eye", "right_eye", "swim_bladder"]
        for i, j in zip(ellipse_shapes, body_parts):
            i.visible = True
            s = series[j]
            i.cx, i.cy, i.a, i.b, i.theta = s.cx, s.cy, s.a, s.b, s.theta

    def _track_img(self, img: np.ndarray) -> np.ndarray:
        return self._track_ellipses(img)

    def _transform_from_roi_to_frame(self, results: np.ndarray):
        if self.roi.value is not None:
            results[:, :2] += self.roi.value[:2]
        return results

    @classmethod
    def _results_to_series(cls, results: np.ndarray):
        eyes_midpoint = results[:2, :2].mean(0)
        swim_bladder_center = results[-1, :2]
        midline = eyes_midpoint - swim_bladder_center
        x2, x1 = midline
        heading = np.rad2deg(np.arctan2(x1, x2))
        theta_l, theta_r = results[:2, -1]
        angle_l = wrap_degrees(theta_l - heading)
        angle_r = wrap_degrees(heading - theta_r)
        s = pd.Series(results.ravel(), index=cls._index)
        s["left_eye", "angle"] = angle_l
        s["right_eye", "angle"] = angle_r
        s["heading"] = heading
        return s



import cv2
import numpy as np

from ztrack.tracking.eye.eye_tracker import EyeParams, EyeTracker
from ztrack.utils.cv import is_in_contour
from ztrack.utils.exception import TrackingError
from ztrack.utils.variable import Float, UInt8


class MultiThresholdEyeTracker(EyeTracker):
    class __Params(EyeParams):
        def __init__(self, params: dict = None):
            super().__init__(params)
            self.sigma = Float("Sigma (px)", 0, 0, 100, 0.1)
            self.threshold_segmentation = UInt8("Segmentation threshold", 127)
            self.threshold_left_eye = UInt8("Left eye threshold", 127)
            self.threshold_right_eye = UInt8("Right eye threshold", 127)
            self.threshold_swim_bladder = UInt8("Swim bladder threshold", 127)

    def __init__(self, roi=None, params: dict = None, *, verbose=0):
        super().__init__(roi, params, verbose=verbose)

    @property
    def _Params(self):
        return self.__Params

    @staticmethod
    def name():
        return "multithreshold"

    @staticmethod
    def display_name():
        return "Multi-threshold"

    def _track_ellipses(self, src: np.ndarray):
        try:
            img = self._preprocess(src, self.params.sigma)

            contours_left_eye = self._binary_segmentation(
                img, self.params.threshold_left_eye
            )
            contours_right_eye = self._binary_segmentation(
                img, self.params.threshold_right_eye
            )
            contours_swim_bladder = self._binary_segmentation(
                img, self.params.threshold_swim_bladder
            )
            contours = self._binary_segmentation(
                img, self.params.threshold_segmentation
            )

            # get the 3 largest contours
            largest3 = sorted(contours, key=cv2.contourArea, reverse=True)[:3]
            assert len(largest3) == 3

            ellipses = self._fit_ellipses(largest3)

            # identify contours
            centers = ellipses[:, :2]
            left_eye, right_eye, swim_bladder = self._sort_centers(centers)

            largest3[swim_bladder] = max(
                contours_swim_bladder,
                key=lambda cnt: is_in_contour(
                    cnt, tuple(centers[swim_bladder])
                ),
            )
            largest3[left_eye] = max(
                contours_left_eye,
                key=lambda cnt: is_in_contour(cnt, tuple(centers[left_eye])),
            )
            largest3[right_eye] = max(
                contours_right_eye,
                key=lambda cnt: is_in_contour(cnt, tuple(centers[right_eye])),
            )

            ellipses = self._fit_ellipses(largest3)
            ellipses = ellipses[[left_eye, right_eye, swim_bladder]]

            return self._correct_orientation(ellipses)
        except (cv2.error, AssertionError):
            raise TrackingError



from __future__ import annotations

from typing import TYPE_CHECKING

from PyQt5 import QtCore, QtWidgets

if TYPE_CHECKING:
    from typing import Iterable, List, Tuple


def selectFiles(filter_: str = None, native=False) -> List[str]:
    dialog = QtWidgets.QFileDialog()
    dialog.setFileMode(QtWidgets.QFileDialog.ExistingFiles)
    dialog.setOption(QtWidgets.QFileDialog.DontUseNativeDialog, not native)

    if filter_ is not None:
        dialog.setNameFilter(filter_)

    return dialog.selectedFiles() if dialog.exec() else []


def selectVideoPaths(
    extensions: Iterable[str] = (".avi", ".mp4"), native=False
):
    filter_ = f'Videos (*{" *".join(extensions)})'
    return selectFiles(filter_, native)


def selectVideoDirectories(
    options: Iterable[Tuple[str, bool]] = None
) -> Tuple[List[str], List[bool]]:
    dialog = QtWidgets.QFileDialog()
    dialog.setWindowFlag(QtCore.Qt.WindowContextHelpButtonHint, False)
    dialog.setWindowFlag(QtCore.Qt.WindowMaximizeButtonHint, True)
    dialog.setFileMode(QtWidgets.QFileDialog.DirectoryOnly)
    dialog.setOption(QtWidgets.QFileDialog.DontUseNativeDialog, True)
    dialog.setMinimumSize(1280, 960)

    checkBoxes: List[QtWidgets.QCheckBox] = []

    if options is not None:
        vBoxLayout = QtWidgets.QVBoxLayout()

        for description, checked in options:
            checkBox = QtWidgets.QCheckBox(dialog)
            checkBox.setText(description)
            checkBox.setChecked(checked)
            vBoxLayout.addWidget(checkBox)
            checkBoxes.append(checkBox)

        groupBox = QtWidgets.QGroupBox(dialog)
        groupBox.setTitle("Options")
        groupBox.setLayout(vBoxLayout)

        layout = dialog.layout()

        if isinstance(layout, QtWidgets.QGridLayout):
            layout.addWidget(groupBox, 4, 0, 1, 3)
        else:
            layout.addWidget(groupBox)

    fileView = dialog.findChild(QtWidgets.QListView, "listView")
    treeView = dialog.findChild(QtWidgets.QTreeView)

    for view in (fileView, treeView):
        if view is not None:
            view.setSelectionMode(
                QtWidgets.QAbstractItemView.ExtendedSelection
            )

    return (dialog.selectedFiles() if dialog.exec() else []), [
        checkBox.isChecked() for checkBox in checkBoxes
    ]



from __future__ import annotations

import webbrowser
from typing import TYPE_CHECKING

from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtWidgets import QSizePolicy, QSpacerItem

from ztrack._run_tracking import run_tracking
from ztrack.gui.create_config import CreateConfigWindow
from ztrack.gui.utils.file import selectVideoDirectories
from ztrack.metadata import homepage, version

from .tracking_viewer import TrackingViewer

if TYPE_CHECKING:
    from typing import Optional

    from ._main_window import MainWindow


class MenuWidget(QtWidgets.QWidget):
    def __init__(self, parent: QtWidgets.QWidget = None, *, verbose=0):
        super().__init__(parent)
        self._verbose = verbose
        self._window: Optional[MainWindow] = None
        self.resize(400, 300)
        self.setWindowTitle("ztrack")

        spacerItem = QSpacerItem(
            40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )
        spacerItem1 = QSpacerItem(
            20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding
        )
        spacerItem2 = QSpacerItem(
            20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding
        )
        spacerItem3 = QSpacerItem(
            40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )

        label = QtWidgets.QLabel(self)
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(label.sizePolicy().hasHeightForWidth())
        label.setSizePolicy(sizePolicy)
        label.setMaximumSize(QtCore.QSize(16777215, 16777215))
        font = QtGui.QFont()
        font.setFamily("Arial")
        font.setPointSize(36)
        label.setFont(font)
        label.setScaledContents(True)
        label.setAlignment(QtCore.Qt.AlignCenter)
        label.setText("ztrack")

        versionLabel = QtWidgets.QLabel(self)
        versionLabel.setText("")
        versionLabel.setAlignment(
            QtCore.Qt.AlignRight  # type: ignore
            | QtCore.Qt.AlignTrailing
            | QtCore.Qt.AlignVCenter
        )

        createConfigPushButton = QtWidgets.QPushButton(self)
        runTrackingPushButton = QtWidgets.QPushButton(self)
        viewResultsPushButton = QtWidgets.QPushButton(self)
        helpPushButton = QtWidgets.QPushButton(self)

        verticalLayout = QtWidgets.QVBoxLayout()
        verticalLayout.addWidget(label)
        verticalLayout.addWidget(versionLabel)
        verticalLayout.addWidget(createConfigPushButton)
        verticalLayout.addWidget(runTrackingPushButton)
        verticalLayout.addWidget(viewResultsPushButton)
        verticalLayout.addWidget(helpPushButton)

        gridLayout = QtWidgets.QGridLayout(self)
        gridLayout.addItem(spacerItem, 1, 0, 1, 1)
        gridLayout.addItem(spacerItem1, 2, 1, 1, 1)
        gridLayout.addItem(spacerItem2, 0, 1, 1, 1)
        gridLayout.addLayout(verticalLayout, 1, 1, 1, 1)
        gridLayout.addItem(spacerItem3, 1, 2, 1, 1)

        createConfigPushButton.setText("Create config")
        runTrackingPushButton.setText("Run tracking")
        viewResultsPushButton.setText("View results")
        helpPushButton.setText("Help")

        versionLabel.setText(f"v{version}")

        createConfigPushButton.clicked.connect(
            self._onCreateConfigPushButtonClicked
        )
        runTrackingPushButton.clicked.connect(
            self._onRunTrackingPushButtonClicked
        )
        viewResultsPushButton.clicked.connect(
            self._onViewResultsPushButtonClicked
        )
        helpPushButton.clicked.connect(lambda: webbrowser.open(homepage))

    @property
    def currentWindow(self) -> Optional[MainWindow]:
        return self._window

    @currentWindow.setter
    def currentWindow(self, value: Optional[MainWindow]):
        if value is None:
            self._window = None
            self.setEnabled(True)
        else:
            self._window = value
            self.setEnabled(False)

    def _onCreateConfigPushButtonClicked(self):
        self.currentWindow = CreateConfigWindow()
        self.currentWindow.closedSignal.connect(self._onWindowClosed)
        self.currentWindow.showMaximized()

    def _onWindowClosed(self):
        self.currentWindow = None

    def _onRunTrackingPushButtonClicked(self):
        inputs, (recursive, overwrite) = selectVideoDirectories(
            (
                (
                    ("Include subdirectories", True),
                    ("Overwrite tracking results", True),
                )
            )
        )
        run_tracking(inputs, recursive, overwrite, self._verbose)

    def _onViewResultsPushButtonClicked(self):
        self.currentWindow = TrackingViewer()
        self.currentWindow.closedSignal.connect(self._onWindowClosed)
        self.currentWindow.showMaximized()


def main(**kwargs):
    from ztrack.gui.utils.launch import launch

    launch(MenuWidget, windowState=QtCore.Qt.WindowNoState, **kwargs)



from functools import reduce

import click

inputs = click.argument("inputs", nargs=-1, type=click.Path(exists=True))
recursive = click.option(
    "-r", "--recursive", is_flag=True, help="Look for files in subdirectories."
)
overwrite = click.option(
    "--overwrite/--no-overwrite", default=True, show_default=True
)
verbose = click.option("-v", "--verbose", count=True, help="Verbosity.")
common_parameters = (inputs, recursive, verbose)


def my_command(f):
    return reduce(lambda x, dec: dec(x), reversed(common_parameters), f)


@click.group()
def main():
    pass


@main.command(short_help="Create tracking configurations for videos.")
@my_command
@click.option(
    "-s",
    "--same-config",
    default=False,
    show_default=True,
    is_flag=True,
    help="Generate the same configuration file for all videos in the "
    "directory",
)
@overwrite
def create_config(**kwargs):
    from ztrack._create_config import create_config

    create_config(**kwargs)


@main.command(
    short_help="Run tracking on videos with created tracking configurations."
)
@my_command
@overwrite
def run(**kwargs):
    from ztrack._run_tracking import run_tracking

    run_tracking(**kwargs)


@main.command(short_help="View tracking results.")
@my_command
def view(**kwargs):
    from ztrack._view_results import view_results

    view_results(**kwargs)


@main.command(short_help="Open GUI.")
@verbose
@click.option("--style", default="dark", show_default=True)
def gui(**kwargs):
    from ztrack.gui.menu_widget import main as main_

    main_(**kwargs)



from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

import pyqtgraph as pg
from PyQt5 import QtCore, QtGui, QtWidgets

from ztrack.utils.shape import Ellipse, Points
from ztrack.utils.variable import Rect

if TYPE_CHECKING:
    from typing import Dict, Iterable, List, Optional

    from ztrack.tracking.tracker import Tracker
    from ztrack.utils.shape import Shape
    from ztrack.utils.typing import config_dict


pg.setConfigOptions(imageAxisOrder="row-major")


class TrackingPlotWidget(pg.PlotWidget):
    roiChanged = QtCore.pyqtSignal(str)
    pointSelected = QtCore.pyqtSignal(int, int)

    def __init__(self, parent: QtWidgets.QWidget = None):
        super().__init__(parent)

        self._imageItem = pg.ImageItem()
        self._rois: Dict[str, Roi] = {}
        self._shapeGroups: Dict[str, List[ShapeGroup]] = {}
        self._currentShapeGroup: Dict[str, ShapeGroup] = {}
        self._currentTab: Optional[str] = None
        self._pointSelectionModeEnabled = False

        self.addItem(self._imageItem)
        self.invertY(True)
        self.setAspectLocked(1)
        self.hideAxis("left")
        self.hideAxis("bottom")
        self.setBackground(None)
        self.scene().sigMouseClicked.connect(self._onMouseClicked)
        self.setRenderHints(QtGui.QPainter.Antialiasing)

    def setPointSelectionModeEnabled(self, enabled):
        self._pointSelectionModeEnabled = enabled
        roi = self._rois[self._currentTab]
        roi.resizable = roi.translatable = not enabled
        self.setCursor(
            QtCore.Qt.CrossCursor if enabled else QtCore.Qt.ArrowCursor
        )

    def _onMouseClicked(self, event):
        if self._pointSelectionModeEnabled:
            pos = event.currentItem.mapSceneToView(event.scenePos())
            self.pointSelected.emit(int(pos.x()), int(pos.y()))

    def setStateFromTrackingConfig(self, trackingConfig: config_dict):
        for groupName, groupDict in trackingConfig.items():
            self._rois[groupName].setRect(groupDict["roi"])

    def setEnabled(self, b: bool):
        for shapeGroup in self._currentShapeGroup.values():
            for shape in shapeGroup.shapes:
                shape.setVisible(b)

        if self._currentTab is not None:
            self._rois[self._currentTab].setVisible(b)

        self._imageItem.setVisible(b)

    def setTracker(self, group_name: str, index: int):
        for shape in self._currentShapeGroup[group_name].shapes:
            self.removeItem(shape)

        self._currentShapeGroup[group_name] = self._shapeGroups[group_name][
            index
        ]

        for shape in self._currentShapeGroup[group_name].shapes:
            self.addItem(shape)
            shape.setVisible(False)
            shape.setBBox(self._rois[group_name].bbox)

    def clearShapes(self):
        for name, shapeGroups in self._shapeGroups.items():
            for shapeGroup in shapeGroups:
                for shape in shapeGroup.shapes:
                    self.removeItem(shape)

    def addTrackerGroup(self, group_name: str, trackers: Iterable[Tracker]):
        roi = self.addRoi(group_name)
        self._shapeGroups[group_name] = [
            ShapeGroup.fromTracker(i) for i in trackers
        ]

        for tracker in trackers:
            tracker.roi = roi.bbox

        roi.sigRegionChanged.connect(lambda: self.roiChanged.emit(group_name))
        self._currentShapeGroup[group_name] = self._shapeGroups[group_name][0]
        self.setTracker(group_name, 0)

    def setTrackerGroup(self, name: str):
        self._setCurrentRoi(name)
        for shape_groups in self._shapeGroups.values():
            for shape_group in shape_groups:
                shape_group.setZValue(1)

        self._currentShapeGroup[name].setZValue(2)

    def setRoiMaxBounds(self, rect):
        for roi in self._rois.values():
            roi.maxBounds = rect

    def setRoiDefaultSize(self, w, h):
        for roi in self._rois.values():
            roi.setDefaultSize(w, h)

    def addRoi(self, name):
        roi = Roi(None, rotatable=False)
        self.addItem(roi)
        roi.setVisible(False)
        self._rois[name] = roi
        return roi

    def _setCurrentRoi(self, name):
        if self._currentTab is not None:
            self._rois[self._currentTab].setVisible(False)

        self._rois[name].setVisible(True)
        self._currentTab = name

    def setImage(self, img):
        self._imageItem.setImage(img)

    def updateRoiGroups(self):
        for shapeGroup in self._currentShapeGroup.values():
            shapeGroup.update()


class Roi(pg.RectROI):
    def __init__(self, bbox=None, **kwargs):
        self._bbox = Rect("", bbox)
        self._default_origin = 0, 0
        self._default_size = 100, 100

        super().__init__(self._pos, self._size, **kwargs)
        self.sigRegionChanged.connect(self._onRegionChanged)

    def setRect(self, rect):
        self._bbox.value = rect
        self.setPos(self._pos)
        self._bbox.value = rect
        self.setSize(self._size)

    @property
    def _pos(self):
        if self._bbox.value is None:
            return self._default_origin

        return self._bbox.value[:2]

    @property
    def _size(self):
        if self._bbox.value is None:
            return self._default_size

        return self._bbox.value[2:]

    def setDefaultSize(self, w, h):
        self._default_size = w, h
        self.setSize(self._size)

    @property
    def bbox(self):
        return self._bbox

    def _onRegionChanged(self):
        x, y = self.pos()
        w, h = self.size()
        self._bbox.value = (x, y, w, h)


class ShapeGroup:
    def __init__(self, shapes: List[pg.ROI]):
        self._shapes = shapes

    def setZValue(self, value):
        for shape in self.shapes:
            shape.setZValue(value)

    @property
    def shapes(self):
        return self._shapes

    @staticmethod
    def fromTracker(tracker: Tracker):
        return ShapeGroup([roiFromShape(shape) for shape in tracker.shapes])

    def update(self):
        for roi in self._shapes:
            roi.refresh()


def roiFromShape(shape: Shape):
    if isinstance(shape, Ellipse):
        return GuiEllipse(shape)
    elif isinstance(shape, Points):
        return GuiPoints(shape)
    else:
        raise NotImplementedError


class ShapeMixin:
    @abstractmethod
    def refresh(self):
        pass

    @abstractmethod
    def setBBox(self, bbox):
        pass


class GuiPoints(pg.ScatterPlotItem, ShapeMixin):
    def __init__(self, points: Points):
        super().__init__()
        self._points = points
        self.setPen(points.lc, width=points.lw)
        self.setSymbol(points.symbol)
        self.refresh()

    def refresh(self):
        if self._points.visible:
            self.setVisible(True)
            self.setData(pos=self._points.data)
        else:
            self.setVisible(False)

    def setBBox(self, bbox):
        self._points.set_bbox(bbox)


class GuiEllipse(QtWidgets.QGraphicsEllipseItem, ShapeMixin):
    def __init__(self, ellipse: Ellipse):
        super().__init__()
        self._ellipse = ellipse
        self.setPen(pg.mkPen(color=ellipse.lc, width=ellipse.lw))
        self.refresh()

    def setBBox(self, bbox):
        self._ellipse.set_bbox(bbox)

    def refresh(self):
        if self._ellipse.visible:
            cx = self._ellipse.cx
            cy = self._ellipse.cy
            a = self._ellipse.a
            b = self._ellipse.b
            theta = self._ellipse.theta
            self.setVisible(True)
            self.setRect(cx - a, cy - b, a * 2, b * 2)
            self.setTransformOriginPoint(cx, cy)
            self.setRotation(theta)
        else:
            self.setVisible(False)



from abc import ABC, abstractmethod
from typing import Union

import numpy as np
import pandas as pd
from matplotlib import cm, colors
from scipy.interpolate import splev, splprep

from ztrack.tracking.params import Params
from ztrack.tracking.tracker import Tracker
from ztrack.utils.shape import Points


class TailParams(Params):
    pass


class TailTracker(Tracker, ABC):
    max_n_points = 99

    def __init__(
        self,
        roi=None,
        params: dict = None,
        *,
        verbose=0,
        cmap: Union[colors.Colormap, str] = "jet",
    ):
        super().__init__(roi, params, verbose=verbose)

        self._tail_cmap = cm.get_cmap(cmap)
        self._points = Points(np.array([[0, 0]]), 1, "m", symbol="+")

    @classmethod
    def _results_to_series(cls, results: np.ndarray):
        n_points = len(results)
        idx = pd.MultiIndex.from_product(
            ((f"point{i:02d}" for i in range(n_points)), ("x", "y"))
        )
        return pd.Series(results.ravel(), idx)

    @abstractmethod
    def _track_tail(self, img: np.ndarray):
        pass

    @staticmethod
    def _interpolate_tail(tail: np.ndarray, n_points: int) -> np.ndarray:
        tck = splprep(tail.T)[0]
        return np.column_stack(splev(np.linspace(0, 1, n_points), tck))

    def _track_img(self, img: np.ndarray):
        tail = self._track_tail(img)
        return self._interpolate_tail(tail, self.params.n_points)

    @property
    def shapes(self):
        return [self._points]

    def annotate_from_series(self, s: pd.Series) -> None:
        tail = s.values.reshape(-1, 2)
        self._points.visible = True
        self._points.data = tail

    def _transform_from_roi_to_frame(self, results: np.ndarray):
        if self.roi.value is not None:
            x0, y0 = self.roi.value[:2]
            results += (x0, y0)
        return results



from PyQt5 import QtCore, QtWidgets
from PyQt5.QtWidgets import QStyle


class FrameBar(QtWidgets.QWidget):
    valueChanged = QtCore.pyqtSignal(int)

    def __init__(
        self, parent: QtWidgets.QWidget = None, *, fps=100, maximum=3000
    ):
        super().__init__(parent)

        self._fps = 0
        self._isPlaying = False

        self._timer = QtCore.QTimer()
        self.setFps(fps)

        self._playIcon = self.style().standardIcon(QStyle.SP_MediaPlay)
        self._pauseIcon = self.style().standardIcon(QStyle.SP_MediaPause)
        self._pushButton = QtWidgets.QPushButton(self)
        self._pushButton.setIcon(self._playIcon)

        self._slider = QtWidgets.QSlider(self)
        self._slider.setOrientation(QtCore.Qt.Horizontal)
        self._spinBox = QtWidgets.QSpinBox(self)

        layout = QtWidgets.QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._pushButton)
        layout.addWidget(self._slider)
        layout.addWidget(self._spinBox)
        self.setLayout(layout)

        self._slider.valueChanged.connect(self._spinBox.setValue)
        self._slider.valueChanged.connect(self.valueChanged.emit)
        self._spinBox.valueChanged.connect(self._slider.setValue)
        self._spinBox.valueChanged.connect(self.valueChanged.emit)

        self._timer.timeout.connect(self._playTick)
        self._pushButton.clicked.connect(self._onPushButtonClicked)

        self.setMaximum(maximum)

    def fps(self):
        return self._fps

    def setFps(self, fps: int):
        self._fps = fps
        self._timer.setInterval(int(1000 / fps))

    def setMaximum(self, value: int):
        self._slider.setMaximum(value)
        self._spinBox.setMaximum(value)

    def value(self):
        return self._slider.value()

    @QtCore.pyqtSlot()
    def _onPushButtonClicked(self):
        self._isPlaying = not self._isPlaying
        if self._isPlaying:
            self._pushButton.setIcon(self._pauseIcon)
            self._timer.start()
        else:
            self._pushButton.setIcon(self._playIcon)
            self._timer.stop()

    @QtCore.pyqtSlot()
    def _playTick(self):
        self._slider.setValue(
            (self._slider.value() + 1) % self._slider.maximum()
        )



from abc import ABC
from typing import List

from ztrack.utils.variable import Variable


class Params(ABC):
    def __setattr__(self, name: str, value):
        if isinstance(value, Variable):
            assert name not in self._parameter_names
            self._parameter_names.append(name)
            self._parameter_list.append(value)
            has_attr = False
            attr = None
            if hasattr(self, name):
                has_attr = True
                attr = object.__getattribute__(self, name)
            object.__setattr__(self, name, value)
            if has_attr:
                self.__setattr__(name, attr)
        elif hasattr(self, name) and isinstance(
            object.__getattribute__(self, name), Variable
        ):
            object.__getattribute__(self, name).value = value
        else:
            object.__setattr__(self, name, value)

    def __getattribute__(self, name: str):
        if isinstance(object.__getattribute__(self, name), Variable):
            return object.__getattribute__(self, name).value
        return object.__getattribute__(self, name)

    def __init__(self, params: dict = None):
        self._parameter_names: List[str] = []
        self._parameter_list: List[Variable] = []
        if params is not None:
            self._set_params(params)

    def _set_params(self, params: dict):
        for key, value in params.items():
            self.__setattr__(key, value)

    @property
    def parameter_list(self):
        return self._parameter_list

    @property
    def parameter_names(self):
        return self._parameter_names

    def to_dict(self):
        return {
            name: value.value
            for name, value in zip(self._parameter_names, self._parameter_list)
        }



import cv2
import numpy as np


def binary_threshold(img: np.ndarray, threshold: int) -> np.ndarray:
    return cv2.threshold(img, threshold, 255, cv2.THRESH_BINARY)[1]


def find_contours(img: np.ndarray) -> list:
    return cv2.findContours(img, cv2.RETR_TREE, cv2.CHAIN_APPROX_NONE)[0]


def is_in_contour(contour: list, point: tuple):
    return cv2.pointPolygonTest(cv2.convexHull(contour), point, False) >= 0



import numpy as np


def wrap_degrees(deg: float) -> float:
    """Constrain angle to [180°, 180°).
    :param deg: Angle in degrees.
    :return: Angle in degrees in [180°, 180°).
    """
    return (deg + 180) % 360 - 180


def angle_diff(a, b):
    diff = (a - b) % (2 * np.pi)
    return np.min((diff, 2 * np.pi - diff), axis=0)



import pandas as pd

from ztrack.tracking import get_trackers_from_config
from ztrack.utils.file import (get_config_dict, get_results_path,
                               get_video_paths_from_inputs)


def run_tracking(
    inputs,
    recursive,
    overwrite,
    verbose,
):
    videos = get_video_paths_from_inputs(inputs, recursive, overwrite)
    for video in videos:
        config = get_config_dict(video)
        trackers = get_trackers_from_config(config, verbose=verbose)
        s = pd.HDFStore(get_results_path(video))
        for key, tracker in trackers.items():
            s[key] = tracker.track_video(video)
        s.close()



from __future__ import annotations

import json
from pathlib import Path
from typing import TYPE_CHECKING

import pandas as pd
from PyQt5 import QtGui, QtWidgets

from ztrack.gui.utils.file import selectVideoDirectories, selectVideoPaths
from ztrack.tracking import get_trackers_from_config
from ztrack.utils.file import (get_config_path, get_paths_for_view_results,
                               get_results_path, video_extensions)

from ._main_window import MainWindow

if TYPE_CHECKING:
    from typing import Dict, List

    from ztrack.tracking.tracker import Tracker


class TrackingViewer(MainWindow):
    def __init__(
        self,
        parent: QtWidgets.QWidget = None,
        videoPaths: List[str] = None,
        verbose=False,
    ):
        super().__init__(parent, videoPaths=videoPaths, verbose=verbose)

        self._results: Dict[str, pd.DataFrame] = {}
        self._trackers: Dict[str, Tracker] = {}
        self.updateVideo()

    def dropEvent(self, event: QtGui.QDropEvent) -> None:
        paths = [u.toLocalFile() for u in event.mimeData().urls()]

        for path in paths:
            self.enqueue(str(path), first=True)

        self.updateVideo()

    def _onFrameChanged(self):
        img = self._currentFrame

        if img is not None:
            self._trackingPlotWidget.setImage(img)

            for name, tracker in self._trackers.items():
                tracker.annotate_from_series(
                    self._results[name].iloc[self._frameBar.value()]
                )
                self._trackingPlotWidget.updateRoiGroups()

    def enqueue(self, videoPath: str, first=False):
        if first:
            self._videoPaths.insert(0, videoPath)
        else:
            self._videoPaths.append(videoPath)

    def dequeue(self):
        if len(self._videoPaths) > 0:
            self._videoPaths.pop(0)
            self._savePaths.pop(0)

    def _openFiles(self):
        videoPaths = selectVideoPaths(native=True)

        for videoPath in reversed(videoPaths):
            self.enqueue(videoPath, first=True)

        self.updateVideo()

    def _openFolders(self):
        directories, (recursive,) = selectVideoDirectories(
            (("Include subdirectories", True),)
        )
        videoPaths = get_paths_for_view_results(
            directories,
            recursive,
        )

        for videoPath in videoPaths:
            self.enqueue(videoPath)

        self.updateVideo()

    def updateVideo(self):
        self._trackingPlotWidget.clearShapes()

        if self._currentVideoPath is not None:
            results_path = get_results_path(self._currentVideoPath)
            config_path = get_config_path(self._currentVideoPath)

            if results_path.exists() and config_path.exists():
                store = pd.HDFStore(results_path)

                for key in store.keys():
                    self._results[key.lstrip("/")] = pd.DataFrame(
                        store.get(key)
                    )

                with open(config_path) as fp:
                    config_dict = json.load(fp)
                self._trackers = get_trackers_from_config(config_dict)

                for name, tracker in self._trackers.items():
                    self._trackingPlotWidget.addTrackerGroup(name, [tracker])
                store.close()

        super().updateVideo()

    def dragEnterEvent(self, event: QtGui.QDragEnterEvent) -> None:
        if event.mimeData().hasUrls():
            paths = [u.toLocalFile() for u in event.mimeData().urls()]

            if all(
                [
                    Path(path).suffix in video_extensions
                    and get_config_path(path).exists()
                    and get_results_path(path).exists()
                    for path in paths
                ]
            ):
                event.accept()
        else:
            event.ignore()



from __future__ import annotations

from typing import TYPE_CHECKING

from PyQt5 import QtCore, QtWidgets

from ztrack.gui.utils.variable_widgets import PointWidget, VariableWidget

if TYPE_CHECKING:
    from typing import Dict, Iterable, List

    from ztrack.gui.create_config import CreateConfigWindow
    from ztrack.tracking.tracker import Tracker
    from ztrack.utils.typing import config_dict, params_dict


class ControlWidget(QtWidgets.QTabWidget):
    trackerChanged = QtCore.pyqtSignal(str, int)
    paramsChanged = QtCore.pyqtSignal(str, int)

    def __init__(self, parent: CreateConfigWindow):
        super().__init__(parent)

        self._trackingPlotWidget = parent.trackingPlotWidget
        self._tabs: Dict[str, TrackingTab] = {}

    @property
    def trackingPlotWidget(self):
        return self._trackingPlotWidget

    def addTrackerGroup(self, groupName: str, trackers: Iterable[Tracker]):
        assert groupName not in self._tabs
        tab = TrackingTab(self, groupName)
        tab.trackerIndexChanged.connect(
            lambda index: self.trackerChanged.emit(groupName, index)
        )

        for tracker in trackers:
            tab.addTracker(tracker)

        self.addTab(tab, groupName.capitalize())
        self._tabs[groupName] = tab

    def getCurrentTrackerIndex(self, groupName: str):
        return self._tabs[groupName].currentIndex

    def setStateFromTrackingConfig(self, trackingConfig: config_dict):
        for groupName, groupDict in trackingConfig.items():
            self._tabs[groupName].setState(
                groupDict["method"], groupDict["params"]
            )


class TrackingTab(QtWidgets.QWidget):
    def __init__(self, parent: ControlWidget, groupName: str):
        super().__init__(parent)

        self._trackingPlotWidget = parent.trackingPlotWidget
        self._parent = parent
        self._groupName = groupName
        self._trackers: List[Tracker] = []
        self._trackerNames: List[str] = []
        self._paramsWidgets: List[ParamsWidget] = []
        self._comboBox = QtWidgets.QComboBox(self)
        self._paramsStackWidget = QtWidgets.QStackedWidget(self)

        label = QtWidgets.QLabel(self)
        label.setText("Method")

        formLayout = QtWidgets.QFormLayout()
        formLayout.setContentsMargins(0, 0, 0, 0)
        formLayout.addRow(label, self._comboBox)

        layout = QtWidgets.QVBoxLayout()
        layout.addLayout(formLayout)
        layout.addWidget(self._paramsStackWidget)
        self.setLayout(layout)

        self.trackerIndexChanged.connect(self.setTracker)

    @property
    def trackingPlotWidget(self):
        return self._trackingPlotWidget

    @property
    def trackerIndexChanged(self) -> QtCore.pyqtBoundSignal:
        return self._comboBox.currentIndexChanged

    @property
    def currentIndex(self):
        return self._comboBox.currentIndex()

    def setTracker(self, i: int):
        self._paramsStackWidget.setCurrentIndex(i)

    def setState(self, methodName: str, params: params_dict):
        index = self._trackerNames.index(methodName)
        self._comboBox.setCurrentIndex(index)
        self._paramsWidgets[index].setParams(params)

    def addTracker(self, tracker: Tracker):
        self._trackerNames.append(tracker.name())
        self._trackers.append(tracker)
        index = self._comboBox.count()
        self._comboBox.addItem(tracker.display_name())
        widget = ParamsWidget(self, tracker=tracker)
        widget.paramsChanged.connect(
            lambda: self._parent.paramsChanged.emit(self._groupName, index)
        )
        self._paramsWidgets.append(widget)
        self._paramsStackWidget.addWidget(widget)


class ParamsWidget(QtWidgets.QFrame):
    paramsChanged = QtCore.pyqtSignal()

    def __init__(self, parent: TrackingTab, *, tracker: Tracker):
        super().__init__(parent)
        self._trackingPlotWidget = parent.trackingPlotWidget
        self._formLayout = QtWidgets.QFormLayout()
        self.setLayout(self._formLayout)
        self._fields: Dict[str, VariableWidget] = {}

        for name, param in zip(
            tracker.params.parameter_names, tracker.params.parameter_list
        ):
            label = QtWidgets.QLabel(self)
            label.setText(param.display_name)
            field = VariableWidget.fromVariable(param, self)

            if isinstance(field, PointWidget):
                field.link(self._trackingPlotWidget)

            field.valueChanged.connect(self.paramsChanged.emit)
            self._fields[name] = field
            self._formLayout.addRow(label, field)

    def setParams(self, params: params_dict):
        for name, value in params.items():
            self._fields[name].setValue(value)



import cv2
import numpy as np

from ztrack.tracking.eye.eye_tracker import EyeParams, EyeTracker
from ztrack.utils.exception import TrackingError
from ztrack.utils.variable import Float, UInt8


class BinaryEyeTracker(EyeTracker):
    def __init__(self, roi=None, params: dict = None, *, verbose=0):
        super().__init__(roi, params, verbose=verbose)

    class __Params(EyeParams):
        def __init__(self, params: dict = None):
            super().__init__(params)
            self.sigma = Float("Sigma (px)", 0, 0, 100, 0.1)
            self.threshold = UInt8("Threshold", 127)

    @property
    def _Params(self):
        return self.__Params

    @staticmethod
    def name():
        return "binary"

    @staticmethod
    def display_name():
        return "Binary threshold"

    def _track_ellipses(self, src: np.ndarray):
        try:
            img = self._preprocess(src, self.params.sigma)

            contours = self._binary_segmentation(img, self.params.threshold)

            # get the 3 largest contours
            largest3 = sorted(contours, key=cv2.contourArea, reverse=True)[:3]
            assert len(largest3) == 3

            # fit ellipses (x, y, semi-major axis, semi-minor axis, theta in
            # degrees)
            ellipses = self._fit_ellipses(largest3)

            centers = ellipses[:, :2]
            ellipses = ellipses[list(self._sort_centers(centers))]

            return self._correct_orientation(ellipses)
        except (cv2.error, AssertionError):
            raise TrackingError



from __future__ import annotations

from abc import ABC
from typing import TYPE_CHECKING

from ztrack.utils.variable import Rect

if TYPE_CHECKING:
    from ztrack.utils.typing import point2d


class Shape(ABC):
    def __init__(self, lw, lc):
        self.lw = lw
        self.lc = lc
        self._bbox = Rect("")
        self._visible = True

    @property
    def visible(self):
        return self._visible

    @visible.setter
    def visible(self, visible: bool):
        self._visible = visible

    def set_bbox(self, bbox):
        self._bbox = bbox

    @property
    def origin(self) -> point2d:
        return 0, 0 if self._bbox.value is None else self._bbox.value[:2]


class Points(Shape):
    def __init__(self, data, lw, lc, symbol):
        super().__init__(lw, lc)
        self._data = data
        self.symbol = symbol

    @property
    def data(self):
        if self._bbox.value is None:
            return self._data
        return self._data + self._bbox.value[:2]

    @data.setter
    def data(self, data):
        self._data = data


class Ellipse(Shape):
    def __init__(
        self, cx: float, cy: float, a: float, b: float, theta: float, lw, lc
    ):
        super().__init__(lw, lc)
        self._cx = cx
        self._cy = cy
        self.a = a
        self.b = b
        self.theta = theta

    @property
    def cx(self):
        if self._bbox.value is None:
            return self._cx
        return self._cx + self._bbox.value[0]

    @cx.setter
    def cx(self, cx: float):
        self._cx = cx

    @property
    def cy(self):
        if self._bbox.value is None:
            return self._cy
        return self._cy + self._bbox.value[1]

    @cy.setter
    def cy(self, cy: float):
        self._cy = cy


class Circle(Ellipse):
    def __init__(self, cx: float, cy: float, r: float, lw, lc):
        super().__init__(cx, cy, r, r, 0, lw, lc)



from __future__ import annotations

from abc import abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING

from decord import VideoReader
from PyQt5 import QtCore, QtWidgets

from ztrack.utils.file import video_extensions

from ._tracking_plot_widget import TrackingPlotWidget
from .utils.frame_bar import FrameBar

if TYPE_CHECKING:
    from typing import List, Optional

    from PyQt5 import QtGui


class MainWindow(QtWidgets.QMainWindow):
    closedSignal = QtCore.pyqtSignal()

    def __init__(
        self,
        parent: QtWidgets.QWidget = None,
        *,
        videoPaths: List[str] = None,
        verbose=False,
    ):
        super().__init__(parent)

        if videoPaths is None:
            videoPaths = []

        self._videoPaths: List[str] = videoPaths
        self._verbose = verbose
        self._frameBar = FrameBar(self)
        self._useVideoFPS = True
        self._videoReader = None
        self._trackingPlotWidget = TrackingPlotWidget(self)
        self._hBoxLayout = QtWidgets.QHBoxLayout()
        self._hBoxLayout.setContentsMargins(0, 0, 0, 0)
        self._hBoxLayout.addWidget(self._trackingPlotWidget)

        self._layout = QtWidgets.QVBoxLayout()
        self._layout.addWidget(self._frameBar)
        self._layout.addLayout(self._hBoxLayout)

        widget = QtWidgets.QWidget(self)
        widget.setLayout(self._layout)
        self.setCentralWidget(widget)

        menuBar = self.menuBar()

        actionOpenFiles = QtWidgets.QAction(self)
        actionOpenFiles.setText("&Open Files")
        actionOpenFiles.setShortcut("Ctrl+O")

        actionOpenFolders = QtWidgets.QAction(self)
        actionOpenFolders.setText("&Open Folders")
        actionOpenFolders.setShortcut("Ctrl+Shift+O")

        actionSetFPS = QtWidgets.QAction(self)
        actionSetFPS.setText("&Set FPS")

        actionAbout = QtWidgets.QAction(self)
        actionAbout.setText("&About")
        actionHelp = QtWidgets.QAction(self)
        actionHelp.setText("&Help")

        fileMenu = menuBar.addMenu("&File")
        viewMenu = menuBar.addMenu("&View")
        helpMenu = menuBar.addMenu("&Help")

        fileMenu.addAction(actionOpenFiles)
        fileMenu.addAction(actionOpenFolders)
        viewMenu.addAction(actionSetFPS)
        helpMenu.addAction(actionAbout)
        helpMenu.addAction(actionHelp)

        actionOpenFiles.triggered.connect(self._openFiles)
        actionOpenFolders.triggered.connect(self._openFolders)
        actionSetFPS.triggered.connect(self._setFPS)
        self._frameBar.valueChanged.connect(self._onFrameChanged)

        self.setMenuBar(menuBar)
        self.setWindowTitle("ztrack")
        self.setAcceptDrops(True)
        self._setEnabled(False)

    @property
    def _currentFrame(self):
        if self._videoReader is None:
            return None

        return self._videoReader[self._frameBar.value()].asnumpy()

    def _setEnabled(self, b: bool):
        self.centralWidget().setEnabled(b)
        self._trackingPlotWidget.setEnabled(b)

    @abstractmethod
    def dropEvent(self, a0: QtGui.QDropEvent) -> None:
        pass

    @property
    def _currentVideoPath(self) -> Optional[str]:
        if len(self._videoPaths) > 0:
            return self._videoPaths[0]

        return None

    @abstractmethod
    def _onFrameChanged(self):
        pass

    @abstractmethod
    def _openFiles(self):
        pass

    @abstractmethod
    def _openFolders(self):
        pass

    def _setFPS(self):
        def onCheckBoxStateChange(state: int):
            isEnabled = state == 0
            spinBox.setEnabled(isEnabled)
            label.setEnabled(isEnabled)

        def onAccepted():
            self._useVideoFPS = checkBox.isChecked()

            if not self._useVideoFPS:
                self._frameBar.setFps(spinBox.value())
            else:
                if self._videoReader is not None:
                    self._frameBar.setFps(int(self._videoReader.get_avg_fps()))

            dialog.close()

        def onRejected():
            dialog.close()

        dialog = QtWidgets.QDialog(None)
        dialog.setWindowFlags(QtCore.Qt.WindowTitleHint)
        dialog.setWindowTitle("Set FPS")
        label = QtWidgets.QLabel(dialog)
        label.setText("FPS")
        spinBox = QtWidgets.QSpinBox(dialog)
        spinBox.setMinimum(0)
        spinBox.setMaximum(1000)
        spinBox.setValue(self._frameBar.fps())

        formLayout = QtWidgets.QFormLayout()
        formLayout.addRow(label, spinBox)

        checkBox = QtWidgets.QCheckBox(dialog)
        checkBox.setText("Use video FPS")

        buttonBox = QtWidgets.QDialogButtonBox(dialog)
        buttonBox.setStandardButtons(
            QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel
        )

        layout = QtWidgets.QVBoxLayout()
        layout.addLayout(formLayout)
        layout.addWidget(checkBox)
        layout.addWidget(buttonBox)
        dialog.setLayout(layout)
        dialog.setMinimumWidth(300)

        checkBox.stateChanged.connect(onCheckBoxStateChange)
        buttonBox.accepted.connect(onAccepted)
        buttonBox.rejected.connect(onRejected)

        checkBox.setChecked(self._useVideoFPS)

        dialog.exec()

    def updateVideo(self):
        if self._currentVideoPath is not None:
            self._videoReader = VideoReader(self._currentVideoPath)
            self._frameBar.setMaximum(len(self._videoReader) - 1)

            if self._useVideoFPS:
                self._frameBar.setFps(int(self._videoReader.get_avg_fps()))

            self._onFrameChanged()
            h, w = self._videoReader[0].shape[:2]
            self._trackingPlotWidget.setRoiDefaultSize(w, h)
            rect = QtCore.QRectF(0, 0, w, h)
            self._trackingPlotWidget.setRoiMaxBounds(rect)
            self._setEnabled(True)
        else:
            self._setEnabled(False)

    def closeEvent(self, a0: QtGui.QCloseEvent):
        self.closedSignal.emit()
        super().closeEvent(a0)

    def dragEnterEvent(self, event: QtGui.QDragEnterEvent) -> None:
        if event.mimeData().hasUrls():
            if all(
                [
                    Path(u.toLocalFile()).suffix in video_extensions
                    for u in event.mimeData().urls()
                ]
            ):
                event.accept()
        else:
            event.ignore()



from ztrack.gui.tracking_viewer import TrackingViewer
from ztrack.gui.utils.launch import launch
from ztrack.utils.file import get_paths_for_view_results


def view_results(
    inputs,
    recursive,
    verbose,
):
    video_paths = [
        str(i) for i in get_paths_for_view_results(inputs, recursive)
    ]
    launch(
        TrackingViewer,
        videoPaths=video_paths,
        verbose=verbose,
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ztrack',
 'ztrack.gui',
 'ztrack.gui.utils',
 'ztrack.tracking',
 'ztrack.tracking.eye',
 'ztrack.tracking.tail',
 'ztrack.utils']

package_data = \
{'': ['*'], 'ztrack.gui': ['img/*']}

install_requires = \
['PyQt5>=5.15.4,<6.0.0',
 'click>=8.0.1,<9.0.0',
 'decord>=0.6.0,<0.7.0',
 'matplotlib>=3.4.2,<4.0.0',
 'opencv-python>=4.5.2,<5.0.0',
 'pandas>=1.3.0,<2.0.0',
 'pyqtgraph>=0.12.2,<0.13.0',
 'qtmodern>=0.2.0,<0.3.0',
 'scikit-image>=0.18.2,<0.19.0',
 'tables>=3.6.1,<4.0.0',
 'tqdm>=4.61.2,<5.0.0']

extras_require = \
{'dev': ['black',
         'isort',
         'flake8',
         'mypy',
         'pydata-sphinx-theme',
         'sphinx',
         'sphinx-autodoc-typehints']}

entry_points = \
{'console_scripts': ['ztrack = ztrack.cli:main']}

setup_kwargs = {
    'name': 'ztrack',
    'version': '0.2.0',
    'description': '',
    'long_description': '[![PyPI version](https://badge.fury.io/py/ztrack.svg)](https://pypi.python.org/pypi/ztrack)\n[![PyPI pyversions](https://img.shields.io/pypi/pyversions/ztrack.svg)](https://pypi.python.org/pypi/ztrack)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Imports: isort](https://img.shields.io/badge/%20imports-isort-%231674b1?style=flat&labelColor=ef8336)](https://pycqa.github.io/isort/)\n[![Checked with mypy](http://www.mypy-lang.org/static/mypy_badge.svg)](http://mypy-lang.org/)\n# ztrack\n\nToolbox for zebrafish pose estimation.\n\n## Installation\n\nUse the package manager [pip](https://pip.pypa.io/en/stable/) to install ztrack.\n\n```bash\npip install ztrack\n```\n',
    'author': 'Ka Chung Lam',
    'author_email': 'kclamar@connect.ust.hk',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/kclamar/ztrack',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

from abc import ABC, ABCMeta, abstractmethod
from typing import TYPE_CHECKING

from PyQt5 import QtCore, QtWidgets

from ztrack.utils.variable import Angle, Float, Int, Point

if TYPE_CHECKING:
    from ztrack.gui._tracking_plot_widget import TrackingPlotWidget
    from ztrack.utils.typing import point2d
    from ztrack.utils.variable import Variable


class AbstractWidgetMeta(type(QtWidgets.QWidget), ABCMeta):  # type: ignore
    pass


class VariableWidget(QtWidgets.QWidget, ABC, metaclass=AbstractWidgetMeta):
    valueChanged = QtCore.pyqtSignal()

    def __init__(
        self, parent: QtWidgets.QWidget = None, *, variable: Variable
    ):
        super().__init__(parent)
        self._variable = variable

    @staticmethod
    def fromVariable(variable: Variable, parent: QtWidgets.QWidget = None):
        if isinstance(variable, Angle):
            return AngleWidget(parent, variable=variable)
        if isinstance(variable, Float):
            return FloatWidget(parent, variable=variable)
        if isinstance(variable, Int):
            return IntWidget(parent, variable=variable)
        if isinstance(variable, Point):
            return PointWidget(parent, variable=variable)
        raise NotImplementedError

    def _setValue(self, value):
        self._variable.value = value
        self.valueChanged.emit()

    @abstractmethod
    def _setGuiValue(self, value):
        pass

    def setValue(self, value):
        self._setValue(value)
        self._setGuiValue(value)


class IntWidget(VariableWidget):
    def __init__(self, parent: QtWidgets.QWidget = None, *, variable: Int):
        super().__init__(parent, variable=variable)

        self._slider = QtWidgets.QSlider(self)
        self._slider.setOrientation(QtCore.Qt.Horizontal)
        self._slider.setValue(variable.value)
        self._slider.setMinimum(variable.minimum)
        self._slider.setMaximum(variable.maximum)
        self._spinBox = QtWidgets.QSpinBox(self)
        self._spinBox.setValue(variable.value)
        self._spinBox.setMinimum(variable.minimum)
        self._spinBox.setMaximum(variable.maximum)

        layout = QtWidgets.QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._slider)
        layout.addWidget(self._spinBox)
        self.setLayout(layout)

        self._slider.valueChanged.connect(self._spinBox.setValue)
        self._spinBox.valueChanged.connect(self._slider.setValue)

        self._slider.valueChanged.connect(self._setValue)
        self._spinBox.valueChanged.connect(self._setValue)

    def _setGuiValue(self, value: int):
        self._spinBox.setValue(value)
        self._slider.setValue(value)


class FloatWidget(VariableWidget):
    def __init__(self, parent: QtWidgets.QWidget = None, *, variable: Float):
        super().__init__(parent, variable=variable)

        self._spinBox = QtWidgets.QDoubleSpinBox(self)
        self._spinBox.setValue(variable.value)
        self._spinBox.setMinimum(variable.minimum)
        self._spinBox.setMaximum(variable.maximum)
        self._spinBox.setSingleStep(variable.step)

        layout = QtWidgets.QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._spinBox)
        self.setLayout(layout)

        self._spinBox.valueChanged.connect(self._setValue)

    def _setGuiValue(self, value: float):
        self._spinBox.setValue(value)


class AngleWidget(VariableWidget):
    class CompassDial(QtWidgets.QDial):
        _valueChanged = QtCore.pyqtSignal(int)

        def __init__(self, parent: QtWidgets.QWidget = None, *, rotation=-90):
            super().__init__(parent)

            self._rotation = rotation
            self.setMinimum(0)
            self.setMaximum(359)
            self.setWrapping(True)
            self.setNotchesVisible(True)
            self.setNotchTarget(90)

            super().valueChanged.connect(
                lambda x: self._valueChanged.emit((x - self._rotation) % 360)
            )

        @property
        def valueChanged(self):
            return self._valueChanged

        def setValue(self, a0: int) -> None:
            super().setValue((a0 + self._rotation) % 360)

    def __init__(
        self,
        parent: QtWidgets.QWidget = None,
        *,
        variable: Angle,
        rotation=-90,
    ):
        super().__init__(parent, variable=variable)

        self._compassDial = AngleWidget.CompassDial(self, rotation=rotation)
        self._spinBox = QtWidgets.QSpinBox(self)
        self._spinBox.setMinimum(0)

        layout = QtWidgets.QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._compassDial)
        layout.addWidget(self._spinBox)
        self.setLayout(layout)

        self._compassDial.setValue(int(variable.value))
        self._spinBox.setMaximum(359)
        self._spinBox.setValue(int(variable.value))
        self._spinBox.setWrapping(True)

        self._compassDial.valueChanged.connect(self._spinBox.setValue)
        self._spinBox.valueChanged.connect(self._compassDial.setValue)

        self._compassDial.valueChanged.connect(self._setValue)
        self._spinBox.valueChanged.connect(self._setValue)

    def _setGuiValue(self, value: int):
        self._spinBox.setValue(value)
        self._compassDial.setValue(value)


class PointWidget(VariableWidget):
    _pointSelectionModeChanged = QtCore.pyqtSignal(bool)

    def __init__(self, parent: QtWidgets.QWidget = None, *, variable: Point):
        super().__init__(parent, variable=variable)

        self._pointSelectionMode = False

        self._pushButton = QtWidgets.QPushButton(self)
        self._pushButton.setText(self._get_display_str(variable.value))

        layout = QtWidgets.QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._pushButton)
        self.setLayout(layout)

        self._pushButton.clicked.connect(
            lambda: self._setPointSelectionMode(not self._pointSelectionMode)
        )

    def _setGuiValue(self, value: point2d):
        self._pushButton.setText(self._get_display_str(value))

    @staticmethod
    def _get_display_str(value: point2d):
        x, y = value
        return f"({x}, {y})"

    def _setPoint(self, x: int, y: int):
        self._setValue((x, y))
        self._setPointSelectionMode(False)

    def _setPointSelectionMode(self, b: bool):
        self._pointSelectionMode = b

        if self._pointSelectionMode:
            self._pushButton.setText("Cancel")
        else:
            self._pushButton.setText(
                self._get_display_str(self._variable.value)
            )

        self._pointSelectionModeChanged.emit(self._pointSelectionMode)

    def link(self, trackingPlotWidget: TrackingPlotWidget):
        self._pointSelectionModeChanged.connect(
            trackingPlotWidget.setPointSelectionModeEnabled
        )
        trackingPlotWidget.pointSelected.connect(self._setPoint)



from abc import ABC, abstractmethod
from typing import Optional, Tuple

import numpy as np

_rect = Optional[Tuple[int, int, int, int]]
_point = Tuple[int, int]


class Variable(ABC):
    def __init__(self, display_name: str):
        self._display_name = display_name

    @property
    def display_name(self):
        return self._display_name

    @property
    @abstractmethod
    def value(self):
        pass

    @value.setter
    def value(self, value):
        pass


class Angle(Variable):
    def __init__(self, display_name: str, angle: float):
        super().__init__(display_name)
        self._value = 0
        self.value = angle

    @property
    def value(self) -> float:
        return self._value

    @value.setter
    def value(self, value: float):
        self._value = int(value) % 360


class Point(Variable):
    def __init__(self, display_name: str, point):
        super().__init__(display_name)

        x, y = point
        point = x, y
        self._value = point

    @property
    def value(self) -> _point:
        return self._value

    @value.setter
    def value(self, value: _point):
        self._value = value


class Rect(Variable):
    def __init__(self, display_name: str, bbox=None):
        super().__init__(display_name)
        if bbox is not None:
            x, y, w, h = bbox
            bbox = x, y, w, h
        self._value = bbox

    @property
    def value(self) -> _rect:
        return self._value

    @value.setter
    def value(self, value: _rect):
        self._value = self._normalize(value)

    @staticmethod
    def _normalize(roi: _rect = None):
        if roi is not None:

            def relu(a):
                return max(0, a)

            x, y, width, height = map(int, roi)
            x0, x1 = sorted(map(relu, (x, x + width)))
            y0, y1 = sorted(map(relu, (y, y + height)))
            return x0, y0, x1 - x0, y1 - y0

    def to_slice(self, axis=0):
        if self._value is None:
            return np.s_[:]
        x, y, width, height = self._value
        return (np.s_[:],) * axis + np.s_[y : y + height, x : x + width]


class Numerical(Variable, ABC):
    def __init__(self, display_name: str, value):
        super().__init__(display_name)
        self._value = value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value


class Bounded(Numerical, ABC):
    def __init__(self, display_name: str, value, minimum, maximum):
        super().__init__(display_name, value)
        assert minimum <= value <= maximum
        self._minimum = minimum
        self._maximum = maximum

    @property
    def value(self):
        return super().value

    @value.setter
    def value(self, value):
        assert self._minimum <= value <= self._maximum
        self._value = value

    @property
    def minimum(self):
        return self._minimum

    @property
    def maximum(self):
        return self._maximum


class Int(Bounded):
    def __init__(
        self, display_name: str, value: int, minimum: int, maximum: int
    ):
        super().__init__(display_name, value, minimum, maximum)


class UInt8(Int):
    def __init__(self, display_name: str, value: int):
        super().__init__(display_name, value, 0, 255)


class Float(Bounded):
    def __init__(
        self,
        display_name: str,
        value: float,
        minimum: float,
        maximum: float,
        step: float,
    ):
        super().__init__(display_name, value, minimum, maximum)
        self._step = step

    @property
    def step(self):
        return self._step



from ztrack.gui.create_config import CreateConfigWindow
from ztrack.gui.utils.launch import launch
from ztrack.utils.file import get_paths_for_config_creation


def create_config(
    inputs,
    recursive,
    same_config,
    overwrite,
    verbose,
):
    video_paths, save_paths = get_paths_for_config_creation(
        inputs, recursive, same_config, overwrite
    )
    launch(
        CreateConfigWindow,
        videoPaths=video_paths,
        savePaths=save_paths,
        verbose=verbose,
    )



import json
from pathlib import Path
from typing import List, Optional

from ztrack._settings import (config_extension, results_extension,
                              video_extensions)


def get_results_path(video):
    return Path(str(video) + results_extension)


def get_config_path(video):
    return Path(str(video) + config_extension)


def get_config_dict(video) -> Optional[dict]:
    path = get_config_path(video)

    if path.exists():
        with open(path) as fp:
            try:
                return json.load(fp)
            except json.JSONDecodeError:
                return None
    return None


def get_video_paths(inputs, recursive):
    paths: List[Path] = list(map(Path, inputs))
    videos = [
        path
        for path in paths
        if path.is_file() and path.suffix in video_extensions
    ]

    for path in filter(Path.is_dir, paths):
        for ext in video_extensions:
            videos.extend(
                path.rglob(f"*{ext}") if recursive else path.glob(f"*{ext}")
            )

    return videos


def get_paths_for_view_results(inputs: List[str], recursive: bool):
    videos = get_video_paths(inputs, recursive)
    videos = [file for file in videos if get_results_path(file).exists()]
    return videos


def get_paths_for_config_creation(
    inputs: List[str], recursive: bool, same_config: bool, overwrite: bool
):
    def _str(path: Path):
        return path.resolve().as_posix()

    paths: List[Path] = [Path(path) for path in inputs]
    directories = [path for path in paths if path.is_dir()]
    files = [
        path
        for path in paths
        if path.is_file() and path.suffix in video_extensions
    ]

    video_paths: List[str] = []
    save_paths: List[List[str]] = []

    for directory in directories:
        for ext in video_extensions:
            if recursive:
                videos = list(directory.rglob(f"*{ext}"))
            else:
                videos = list(directory.glob(f"*{ext}"))

            if not overwrite:
                videos = [
                    video
                    for video in videos
                    if not video.with_suffix(results_extension).exists()
                ]

            if len(videos) > 0:
                if same_config:
                    video_paths.append(_str(videos[0]))
                    save_paths.append([_str(video) for video in videos])
                else:
                    for video in videos:
                        video_paths.append(_str(video))
                        save_paths.append([_str(video)])

    if not overwrite:
        files = [
            file
            for file in files
            if not file.with_suffix(results_extension).exists()
        ]

    for file in files:
        video_paths.append(_str(file))
        save_paths.append([_str(file)])

    return video_paths, save_paths


def get_video_paths_from_inputs(
    inputs: List[str], recursive: bool, overwrite: bool
):
    videos = get_video_paths(inputs, recursive)
    videos = [file for file in videos if get_config_path(file).exists()]

    if not overwrite:
        videos = [
            file for file in videos if not get_results_path(file).exists()
        ]

    return videos
